#include  "buffer.h"
#include <vector>
#include  <cstring>
#include  <iostream>
#include  "global.h" 
#include  <cctype>
#include  <cstdlib>
#include  <algorithm> 
#include  <cassert> 

namespace atlas
{ namespace interpreter
  {

String_pool::String_pool(size_t b)
: block_size(b),start(nullptr),point(nullptr),chars_left(0),prev(nullptr)
{}

char* String_pool::store(const char* s,size_t len)
{ size_t n= std::max(len+1,block_size); // size of new block if needed
  if (len>=chars_left) // then the current block cannot store |s|, so allocate
  { if (start!=nullptr) // initial block needs no backing up
      prev=new backup(start,prev); // push pointers to chain, for destruction
    start=point=new char[n]; // now we can copy to where |point| points
    chars_left=n;
  }
  char* result=strncpy(point,s,len);
  point+=len;  *point++='\0'; // write |'\0'| whether or not present in |s|
  chars_left-=len+1;
  return result;
}

String_pool::backup::~backup() { delete[] block_start; delete link; }

const id_type Hash_table::empty = ~static_cast<id_type>(0);

Hash_table::Hash_table(size_t init,size_t block_size)
: pool(block_size)
, mod(init<2?2:init), hash_tab(mod,empty), name_tab(0)
{ name_tab.reserve(max_fill());
}

id_type Hash_table::hash (const char* s, size_t l) const
  // |s| a string of length |l>0|
{ unsigned long r=static_cast<unsigned char>(*s++)%mod;
  while (l-->1) // one less, because we already read one character
    r=(static_cast<unsigned char>(*s++)+(r<<8))%mod;
  return static_cast<id_type>(r);
}

bool Hash_table::knows (const char* name) const
{ size_t l = std::strlen(name);
  id_type i,h=hash(name,l);
  while ((i=hash_tab[h])!=empty) // search breaks off at first empty slot
    if (std::strncmp(name_tab[i],name,l)==0 && name_tab[i][l]=='\0')
      return true; // identifier found in table
    else
      if (++h==mod) // move past occupied slot
        h=0; // possibly wrapping around
  return false;
}

id_type Hash_table::do_match (const char* str, size_t l, bool copy_string)
{ id_type i,h=hash(str,l);
  while ((i=hash_tab[h])!=empty)
    if (std::strncmp(name_tab[i],str,l)==0 && name_tab[i][l]=='\0')
      return i; /* identifier found in table */
    else if (++h==mod)
      h=0; /* move past occupied slot */
  if (name_tab.size()>=max_fill())
  { 
    { std::vector<id_type>(mod=2*mod-1,empty).swap(hash_tab); // make length odd
      for (size_t i=0; i<name_tab.size(); ++i)
      { int h=hash(name_tab[i],std::strlen(name_tab[i])); // rehash the string
        while (hash_tab[h]!=empty) 
          if (++h==mod)
            h=0; /* find empty slot */
        hash_tab[h]=i;
      }
      name_tab.reserve(max_fill());
    }
    h=hash(str,l); 
    while (hash_tab[h]!=empty) 
      if (++h==mod) h=0; /* find free slot */
  }
  hash_tab[h]= name_tab.size(); /* this number represents the new identifier */
  name_tab.push_back(copy_string ? pool.store(str,l) : str);
  return hash_tab[h];
}

BufferedInput::BufferedInput (std::istream& s)
:base_stream(s)
,line_buffer()
,pos(nullptr)
,prompt(nullptr)
,prompt2(nullptr)
,def_ext(nullptr)
,temp_prompt(nullptr)
,readline(nullptr)
,add_hist(nullptr)
,line_no(1)
,cur_lines(0)
,input_stack()
,input_files_seen()
,input_files_completed()
,stream(&base_stream)
{}

BufferedInput::BufferedInput
(const char* pr, rl_type rl, add_hist_type ah,const char* pr2,const char* de)
:base_stream(std::cin)
,line_buffer()
,pos(nullptr)
,prompt(pr)
,prompt2(pr2)
,def_ext(de)
,temp_prompt("")
,readline(rl)
,add_hist(ah)
,line_no(1)
,cur_lines(0)
,input_stack()
,input_files_seen()
,input_files_completed()
,stream(&base_stream)
{}

void BufferedInput::pop_file()
{ line_no = input_stack.top().line_no;
  *output_stream << "Completely read file '" << cur_fname() << "'."
                 << std::endl;
  input_files_completed.insert (input_stack.top().name); // reading succeeded
  input_stack.pop(); // also closes the current file
  stream= input_stack.empty() ? &base_stream : &input_stack.top().f_stream;
}

void BufferedInput::close_includes()
{ while (not input_stack.empty())
  { std::cerr << "Abandoning reading of file '" << cur_fname() 
              << "' at line " << line_no << std::endl;
  line_no = input_stack.top().line_no;
    input_stack.pop();
  }
  stream= &base_stream;
}

BufferedInput::input_record::input_record
(BufferedInput& parent, const char* file_name)
: f_stream()
  , name(~0)
  , line_no(parent.line_no+parent.cur_lines) // record where reading will resume

{ unsigned int path_size = input_path_size();
  for (unsigned int i=0; i<=path_size; ++i)
  { std::string pathname(i<path_size ? input_path_component(i) : std::string());
    pathname += file_name;
    f_stream.open(pathname);
    if (not f_stream.is_open() and parent.def_ext!=nullptr)
    { long inx=pathname.size()-std::strlen(parent.def_ext);
      if (inx<0 or pathname.substr(inx)!=parent.def_ext)
        // only add extension if absent
        f_stream.open(pathname+= parent.def_ext);
          // try to reopen |stream| with extended name
    }
    if (f_stream.is_open())
       // once opening succeeds, record name in |input_files_seen|
    {name = parent.input_files_seen.match(pathname.c_str(),pathname.size());
      return;
    }
  }
}

bool BufferedInput::push_file(const char* name, bool skip_seen)
{
  if (skip_seen and input_files_seen.knows(name) 
      and input_files_completed.isMember(input_files_seen.match_literal(name)))
    return true;
  input_stack.emplace(*this,name); // this tries to open the file

  if (input_stack.top().f_stream.is_open())
  { 
    { bool skip=false;
      const id_type file_nr = input_stack.top().name;
      if (file_nr==input_files_completed.capacity()) // it wasn't seen before
        input_files_completed.extend_capacity(false);
           // add new empty slot to bitmap; may set it later
      else // old name; need to do some checks
      { skip = skip_seen and input_files_completed.isMember(file_nr);
        auto it=input_stack.ctop();
        while (not skip and ++it!=input_stack.cbottom())
          if (file_nr==it->name)
            skip=true; // avoid recursive inclusion of active file
      }
      if (skip)
        input_stack.pop(); // silently skip including the file
      else
      { *output_stream << "Starting to read from file '" << cur_fname()
                       << "'." << std::endl;
      stream= &input_stack.top().f_stream;
        line_no=1; // prepare to read from pushed file
        cur_lines=0;
          // so we won't advance |line_no| when getting first line of new file
      }
    }
    return true; // succeed whether or not a file was actually pushed
  }
  else // opening file failed
  { input_stack.pop();
    // don't call |pop_file|: |stream| and |line_no| were not updated
    std::cerr << "failed to open input file '" << name << "'." << std::endl;
    return false;
  }
}

bool BufferedInput::getline()
{ line_buffer="";
  line_no+=cur_lines; cur_lines=0;
  std::string pr=
                (temp_prompt.empty() ? std::string(prompt==nullptr?"":prompt)
                                     : temp_prompt+" "+prompt2)
;
  bool go_on, popped=false; unsigned int size=0;
  containers::simple_list<std::string> lines;
  do
  { lines.emplace_front(); // allocate a new empty |std::string|
    auto& line = lines.front();
  
  { if (stream->good())
      { 
        if (input_stack.empty() and prompt!=nullptr)
          // do only at top level, and only if prompt enabled
        { prompt_length= pr.length();
          if (readline!=nullptr) // skip calling 'readline' if no function is supplied
          { char* l=readline(pr.c_str());
            if (l==nullptr) // then |readline| failed, flag end of file
            { line=""; stream->setstate(std::ios_base::eofbit);
              std::cout << "^D\n";
            }
            else
            { line=l;
              if (add_hist!=nullptr and *l!='\0')
                add_hist(l);
              else
                std::free(l);
            }
          }
          else
          { std::cout << pr;
              std::getline(std::cin,line,'\n'); }
        }
        else
           std::getline(*stream,line,'\n');
   }
    if (line.empty() and not stream->good()) // nothing found and file ended
    { if (cur_lines>0)
      { std::cerr << "end of file after backslash ";
        if (input_stack.empty())
          std::cerr << (stream==&std::cin ? "on standard input" : "in main file")
                    << std::endl;
        else
        { std::cerr << "in file '" << cur_fname() << '\'' << std::endl;
        pop_file(); popped=true;
        }
      }
      else if (input_stack.empty())
        return false; // all hope of returning something has gone
      else { pop_file(); popped=true; }
      break;
  // in these cases terminate line-continuation; return the part that was found
    }
    else {} // something (at least a bona fide empty line) was contributed here
  }
    ++cur_lines;
    auto l=line.length();
    while (l>0 and std::isspace(line[l-1]))
      --l; // back up over trailing space
    go_on=(l>0 and line[l-1]=='\\'); // keep going only if final backslash present
    if (go_on)
    { pr= "\\ ";
      --l;
    } // set prompt and chop off blackslash
    size+=l;
    line.erase(l); // truncate line
  }
  while(go_on);
  
  { line_buffer.resize(size+1);
    pos=&*line_buffer.end();
    *--pos = popped ? '\f' :'\n'; // terminate with newline or end-of-file
    for (auto it=lines.begin(); not lines.at_end(it); ++it)
      // concatenate |lines|, reversing order
      pos=std::copy_backward(it->begin(),it->end(),pos);
    assert(pos==line_buffer.data()); // we should have come to the start
  }
return true;
  // delay reporting end of input if anything was read at all
}

void BufferedInput::locate (const char* p, int& line, int& column) const
{ line=line_no; column=p-line_buffer.data(); }

void BufferedInput::show_range
  (std::ostream& out,unsigned long l0, int c0, unsigned long l1, int c1)
  const
{ if (l1==line_no)
  { int pl=prompt_length; // offset of last line on the screen
    if (stream!=&std::cin or cur_lines>1)
    { if (not input_stack.empty())
        out << "In input file '" << cur_fname() << "', ";
      if (stream!=&std::cin)
        out << "line " << l0 << ":\n";
      out<<line_buffer; pl=0; // echo line in these cases
    }
    if (l0<l1)
      c0=0; // forget start column of multi-line range
    for (int i=pl+c0; i>0; --i)
      out<<' ';
    for (int i=c1; i>c0; --i)
      out<<'^';
    out << std::endl;
    if (l0<l1)
      if (l1-l0==1)
        out << "Range started in previous line\n";
      else
        out << "Range started " << (l1-l0) << "lines above\n";
  }
  else // range ended before the current line
    out << "Range from line " << l0 << " column " << c0 
	<< " to line " << l1 << " column " << c1 << ".\n";
}

void BufferedInput::push_prompt(char c) 
{ temp_prompt.push_back(c); }

char BufferedInput::top_prompt() const
{ auto l=temp_prompt.length();
  return l==0 ? '\0' : temp_prompt[l-1];
}

void BufferedInput::pop_prompt()
{ auto l=temp_prompt.length();
  if (l>0)
    temp_prompt.erase(l-1);
}

void BufferedInput::reset() 
{ temp_prompt=""; }


BufferedInput* main_input_buffer=nullptr;

  }
}

