#include  "global.h"
#include  <cstdlib>
#include  "lexer.h" 
#include  "axis.h"
#include  "parsetree.h" 
#include <sstream>
#include <iomanip>
#include  "matreduc.h"
#include  "bitmap.h"
#include  "lattice.h"
#include  "bitvector.h"
#include  "permutations.h"

namespace atlas { namespace interpreter {

Id_table* global_id_table=nullptr; 

overload_table* global_overload_table=nullptr;

std::ostream* output_stream= &std::cout;

bool clean=true;

int verbosity=0;

const shared_bool global_false = std::make_shared<bool_value>(false);
const shared_bool global_true  = std::make_shared<bool_value>(true);

namespace {

std::size_t locate_overload
  (id_type id,const overload_table::variant_list& slot,const type_expr& arg_type)
{ std::size_t lwb=0; std::size_t upb=slot.size();
  for (std::size_t i=0; i<slot.size(); ++i)
  { unsigned int cmp= is_close(arg_type,slot[i].type().arg_type);
    switch (cmp)
    {
      case 0x6: lwb=i+1; break;
        // existent type |i| converts to |type|, which must come later
      case 0x5:  if (upb>i) upb=i;  break;
        // |type| converts to type |i|, so it must come before
      case 0x7: // mutually convertible types, maybe identical ones
        if (slot[i].type().arg_type==arg_type)
          // identical ones: overload redefinition case
            return i;
      // |else| {\bf fall through}
      case 0x4:
       
         throw program_error()
           << "Cannot overload `" << main_hash_table->name_of(id) << "':\n" 
              "already overloaded type '" << slot[i].type().arg_type
         << "' is too close to new argument type '" << arg_type
         << "',\nwhich would make overloading ambiguous for certain arguments. " 
              "Simultaneous\noverloading for these types is not possible, " 
              "forget the other one first.";
      default: {} // nothing for unrelated argument types
    }
  }
  if (lwb>upb)
    throw logic_error("Conflicting order of related overload types");

  return upb;
}


void add_overload(id_type id, shared_function&& f, type_expr&& type)
{
  auto old_n=global_overload_table->variants(id).size();
std::ostringstream type_string;
  type_string << type;
    // save type |type| as string before moving from it
  global_overload_table->add(id,std::move(f),std::move(type));
    // insert or replace table entry
  auto n=global_overload_table->variants(id).size();
  if (n==old_n)
    *output_stream << "Redefined ";
  else if (n==1)
    *output_stream << "Defined ";
  else
    *output_stream << "Added definition [" << n << "] of ";
  *output_stream << main_hash_table->name_of(id) << ": "
            << type_string.str();
  *output_stream << std::endl;
}

void handle
  (const std::exception& err,const id_pat& pat, int phase, int overload,
   const source_location& loc)
{ static const char* message[3] = {"not executed","interrupted","failed"};
  std::cerr << "Error in 'set' command " << loc << ":\n" 
            << err.what() << "\n  Command 'set " << pat << "' "
            << message[phase];
  if (phase<2)
    std::cerr << ", nothing " << (overload<=1 ? "defin" : "overload") << "ed";
  std::cerr << ".\n";
clean=false;
  reset_evaluator(); main_input_buffer->close_includes();
}
void do_global_set(id_pat&& pat, const expr& rhs, int overload,
                  const source_location& loc)
{ auto n_id=count_identifiers(pat);
  int phase; // needs to be declared outside the |try|, is used in |catch|
  try
  { phase=0; // type check
    expression_ptr e;
    type_expr t=analyse_types(rhs,e);
    if (not pattern_type(pat).specialise(t))
      
      throw program_error()
          << "Type " << t 
          << " of right hand side does not match required pattern "
          << pattern_type(pat);
    
    { if (overload==2 and t.kind()!=function_type)
        throw program_error() << "Cannot set operator '" << pat.name 
           << "' to a value of non-function type " << t;
    }
    definition_group b(n_id);
    b.thread_bindings(pat,t); // match identifiers and their future types

    phase=1; // evaluation of right hand side
  e->eval();

    phase=2; // actual definition of identifiers
    std::vector<shared_value> v;
    v.reserve(n_id);
    thread_components(pat,pop_value(),std::back_inserter(v));
     // associate values with identifiers
    auto v_it = v.begin();
    for (auto it = b.begin(); it!=b.end(); ++it,++v_it)
    { assert(v_it!=v.end());
      
      { unsigned int input_level = main_input_buffer->include_depth();
        *output_stream << std::setw(2*input_level) << "";
      }
      if (overload==0 or it->second.kind()!=function_type)
      
      { *output_stream << (b.is_const(it) ? "Constant " : "Variable ") 
                       << main_hash_table->name_of(it->first)
                       << ": " << it->second;
        if (global_id_table->present(it->first))
        { bool is_const;
          *output_stream << " (overriding previous instance, which had type "
                    << *global_id_table->type_of(it->first,is_const);
          if (is_const)
            *output_stream << " (constant)";
          *output_stream << ')';
        }
        *output_stream << std::endl;
        global_id_table->add
          (it->first,std::move(*v_it),std::move(it->second),b.is_const(it));
      }
      else
      
      { shared_function f = std::dynamic_pointer_cast<const function_base>(*v_it);
        if (f.get()==nullptr)
          throw logic_error("Non-function value found with function type");
        add_overload(it->first,std::move(f),std::move(it->second));
      }
    }
  }
  
  catch (const program_error& err)
  { assert(phase!=1); handle(err,pat,phase,overload,loc); }
  catch (const runtime_error& err)
  { assert(phase==1); handle(err,pat,phase,overload,loc); }
  catch (const logic_error& err)
  { std::cerr << "Unexpected error: ";
    handle(err,pat,phase,overload,loc);
  }
  catch (const std::exception& err)
  { handle(err,pat,phase,overload,loc); }
}

void rational_convert() // convert integer to rational (with denominator~1)
{ own_int i = get_own<int_value>();
  push_value(std::make_shared<rat_value>(big_rat(std::move(i->val))));
}

void intlist_vector_convert()
{ shared_row r = get<row_value>();
  push_value(std::make_shared<vector_value>(row_to_vector(*r)));
}

void ratlist_ratvec_convert() // convert list of rationals to rational vector
{ shared_row r = get<row_value>();
  big_int d(1);
  Ratvec_Numer_t numer(r->val.size());
    // type |Ratvec_Numer_t| is needed near end
  std::vector<arithmetic::Denom_t> denom(r->val.size());
  for (std::size_t i=0; i<r->val.size(); ++i)
  // collect numerators and denominators separately
  { const auto* frac = force<rat_value>(r->val[i].get());
    numer[i]=frac->numerator().long_val();
    denom[i]=frac->denominator().ulong_val();
    d=lcm(d,frac->denominator()); // compute least common denominator safely
  }
  for (std::size_t i=0; i<r->val.size(); ++i)
  { big_int n =
      big_int::from_signed(numer[i])*(d/big_int::from_unsigned(denom[i]));
    numer[i] = n.long_val();
    // adjust numerators to common denominator, if it fits
  }

  push_value(std::make_shared<rational_vector_value>
     (std::move(numer),d.ulong_val()));
}

void vector_intlist_convert()
{ shared_vector v = get<vector_value>();
  push_value(vector_to_row(v->val));
}

void ratvec_ratlist_convert() // convert rational vector to list of rationals
{ shared_rational_vector rv = get<rational_vector_value>();
  own_row result = std::make_shared<row_value>(rv->val.size());
  for (std::size_t i=0; i<rv->val.size(); ++i)
    result->val[i] = std::make_shared<rat_value>
      (Rational(rv->val.numerator()[i],rv->val.denominator()));
  push_value(result);
}

rat_Vector introw_to_ratvec(const row_value& r)
{ Ratvec_Numer_t numer(r.val.size());
  for(std::size_t i=0; i<r.val.size(); ++i)
    numer[i]=force<int_value>(r.val[i].get())->val.long_val();
  return rat_Vector(numer,1);
}
own_row vector_to_ratrow(const int_Vector& v)
{ own_row result = std::make_shared<row_value>(v.size());
  for(std::size_t i=0; i<v.size(); ++i)
    result->val[i]=std::make_shared<rat_value>(big_rat(big_int(v[i])));
  return result;
}
own_row introw_to_ratrow(own_row r)
{ for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it=std::make_shared<rat_value>
       (big_rat(std::move(force<int_value>(it->get())->val)));
  return r;
}

void vector_ratvec_convert() // convert vector to rational vector
{ shared_vector v = get<vector_value>();
  push_value(std::make_shared<rational_vector_value> (rat_Vector(v->val,1)));
}

void intlist_ratvec_convert()
{ shared_row r = get<row_value>();
  push_value(std::make_shared<rational_vector_value>
    (introw_to_ratvec(*r)));
}

void vector_ratlist_convert()
{ shared_vector v = get<vector_value>();
  push_value(vector_to_ratrow(v->val));
}

void intlist_ratlist_convert()
{ push_value(introw_to_ratrow(get_own<row_value>()));
}

void veclist_matrix_convert()
{ shared_row r = get<row_value>();
  if (r->val.size()==0)
    throw runtime_error
      ("Implicit conversion to matrix for an empty set of vectors");

  auto n = force<vector_value>(r->val[0].get())->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(std::size_t j=0; j<r->val.size(); ++j)
  { const int_Vector& col = force<vector_value>(r->val[j].get())->val;
    if (col.size()!=n)
      throw runtime_error("Vector sizes differ in conversion to matrix");

    m->val.set_column(j,col);
  }
  push_value(std::move(m));
}

void intlistlist_matrix_convert()
{ shared_row r = get<row_value>();
  if (r->val.size()==0)
    throw runtime_error("Cannot convert empty list of lists to matrix");

  auto n = force<row_value>(r->val[0].get())->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(std::size_t j=0; j<r->val.size(); ++j)
  { int_Vector col = row_to_vector(*force<row_value>(r->val[j].get()));
    if (col.size()!=n)
      throw runtime_error("List sizes differ in conversion to matrix");

    m->val.set_column(j,col);
  }
  push_value(std::move(m));
}

void intlistlist_veclist_convert()
{ own_row r = get_own<row_value>();
  for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it = std::make_shared<vector_value> 
      (row_to_vector(*force<row_value>(it->get())));
  push_value(r);
}

void matrix_veclist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(unsigned int j=0; j<m->val.numColumns(); ++j)
    result->val[j]=std::make_shared<vector_value>(m->val.column(j));
  push_value(result);
}

void matrix_intlistlist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(unsigned int j=0; j<m->val.numColumns(); ++j)
    result->val[j]= vector_to_row(m->val.column(j));
  push_value(result);
}

void veclist_intlistlist_convert()
{ own_row r=get_own<row_value>();
  for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it = vector_to_row(force<vector_value>(it->get())->val);
  push_value(r);
}

void matrix_ratveclist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(unsigned int j=0; j<m->val.numColumns(); ++j)
    result->val[j]=std::make_shared<rational_vector_value> 
      (rat_Vector(m->val.column(j),1));
  push_value(result);
}

void matrix_ratlistlist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(unsigned int j=0; j<m->val.numColumns(); ++j)
  result->val[j]= vector_to_ratrow(m->val.column(j));
  push_value(result);
}

void veclist_ratveclist_convert()
{ own_row r=get_own<row_value>();
  for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it=std::make_shared<rational_vector_value> 
      (rat_Vector(force<vector_value>(it->get())->val,1));
  push_value(r);
}

void veclist_ratlistlist_convert()
{ own_row r=get_own<row_value>();
  for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it = vector_to_ratrow(force<vector_value>(it->get())->val);
  push_value(r);
}


void intlistlist_ratveclist_convert()
{ own_row r=get_own<row_value>();
  for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it=std::make_shared<rational_vector_value> 
       (introw_to_ratvec(*force<row_value>(it->get())));
  push_value(r);
}

void intlistlist_ratlistlist_convert()
{ own_row r=get_own<row_value>();
  for (auto it=r->val.begin(); it!=r->val.end(); ++it)
    *it=introw_to_ratrow(force_own<row_value>(std::move(*it)));
  push_value(r);
}

void plus_wrapper(expression_base::level l)
{ own_int j=get_own<int_value>();
  shared_int i=get<int_value>(); // |j| more likely |unique|
  if (l==expression_base::no_value)
    return;
  j->val += i->val;
  push_value(j);
}

void minus_wrapper(expression_base::level l)
{ own_int j=get_own<int_value>();
  shared_int i=get<int_value>(); // |j| more likely |unique|
  if (l==expression_base::no_value)
    return;
  j->val.subtract_from(i->val);
  push_value(j);
}

void times_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l==expression_base::no_value)
    return;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i->val*j->val));
}

void divide_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  own_int i=get_own<int_value>();
  if (j->val.is_zero())
    throw runtime_error("Division by zero");
  if (l==expression_base::no_value)
    return;
  i->val /= j->val;
  push_value(i);
}

void modulo_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  own_int i=get_own<int_value>();
  if (j->val.is_zero())
    throw runtime_error("Modulo zero");
  if (l==expression_base::no_value)
    return;
  i->val %= j->val;
  push_value(i);
}

void divmod_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  own_int i=get_own<int_value>();
  if (j->val.is_zero())
    throw runtime_error("DivMod by zero");
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<int_value>(i->val.reduce_mod(j->val)));
  // quotient
  push_value(i); // remainder
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void unary_minus_wrapper(expression_base::level l)
{ own_int i=get_own<int_value>();
  if (l!=expression_base::no_value)
    i->val.negate(), push_value(i);
}

void power_wrapper(expression_base::level l)
{ static shared_int one = std::make_shared<int_value>(1);
  // constants shared between calls
  static shared_int minus_one  = std::make_shared<int_value>(-1);

  int n=get<int_value>()->int_val(); // exponent is small
  shared_int b=get<int_value>(); // base can be large
  bool unit_base = b->val.size()==1 and std::abs(b->int_val())==1;
  if (n<0 and not unit_base)
    throw runtime_error("Negative power of integer");
  if (l==expression_base::no_value)
    return;

  if (unit_base)
  { push_value(n%2!=0 and b->val.is_negative() ? minus_one : one);
      return;
  }

  push_value(std::make_shared<int_value>(b->val.power(n)));
}

void fraction_wrapper(expression_base::level l)
{ shared_int d=get<int_value>();
  shared_int n=get<int_value>();
  if (d->val.is_zero())
    throw runtime_error("fraction with zero denominator");
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<rat_value>
     (big_rat::from_fraction(n->val,d->val)));
}


void unfraction_wrapper(expression_base::level l)
{ own_rat q=get_own<rat_value>();
  if (l!=expression_base::no_value)
  { push_value(std::make_shared<int_value>(std::move(q->numerator())));
    push_value(std::make_shared<int_value>(std::move(q->denominator())));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void rat_plus_int_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  own_rat q=get_own<rat_value>();
  if (l!=expression_base::no_value)
  {
    q->val+=i->val;
    push_value(q);
  }
}
void rat_minus_int_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  own_rat q=get_own<rat_value>();
  if (l!=expression_base::no_value)
  {
    q->val-=i->val;
    push_value(q);
  }
}
void rat_times_int_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  shared_rat q=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(q->val*i->val));
}
void rat_divide_int_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  shared_rat q=get<rat_value>();
  if (i==0)
    throw runtime_error("Rational division by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(q->val/i->val));
}

void rat_quotient_int_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  shared_rat q=get<rat_value>();
  if (i->val.is_zero())
    throw runtime_error("Rational quotient by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(q->val.quotient(i->val)));
}

void rat_modulo_int_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  own_rat q=get_own<rat_value>();
  if (i==0)
    throw runtime_error("Rational modulo zero");
  if (l!=expression_base::no_value)
  {
    q->val%=i->val;
    push_value(q);
  }
}

void rat_plus_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>();
  shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val+j->val));
}
void rat_minus_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>();
  shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val-j->val));
}
void rat_times_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>();
  shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val*j->val));
}
void rat_divide_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>();
  shared_rat i=get<rat_value>();
  if (j->val.numerator()==0)
    throw runtime_error("Rational division by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val/j->val));
}
void rat_modulo_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>();
  shared_rat i=get<rat_value>();
  if (j->val.numerator()==0)
    throw runtime_error("Rational modulo zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val%j->val));
}

void rat_unary_minus_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(-i->val));
}
void rat_inverse_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (i->val.numerator()==0)
    throw runtime_error("Inverse of zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val.inverse()));
}


void rat_floor_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i->val.floor()));
}
void rat_ceil_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i->val.ceil()));
}
void rat_frac_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i->val.frac()));
}
void rat_power_wrapper(expression_base::level l)
{ int n=get<int_value>()->int_val(); own_rat b=get_own<rat_value>();
  if (b->val.numerator()==0 and n<0)
    throw runtime_error("Negative power of zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(b->val.power(n)));
}

void int_unary_eq_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_zero()));
}
void int_unary_neq_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not i->val.is_zero()));
}
void int_non_negative_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not i->val.is_negative()));
}
void int_positive_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not(i->val.is_negative() or i->val.is_zero())));
}
void int_non_positive_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_negative() or i->val.is_zero()));
}
void int_negative_wrapper(expression_base::level l)
{ shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_negative()));
}

void int_eq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}

void int_neq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void int_less_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<j->val));
}

void int_lesseq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<=j->val));
}

void int_greater_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>j->val));
}

void int_greatereq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>=j->val));
}

void rat_unary_eq_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_zero()));
}
void rat_unary_neq_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not i->val.is_zero()));
}
void rat_non_negative_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not i->val.is_negative()));
}
void rat_positive_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not (i->val.is_negative() or i->val.is_zero())));
}
void rat_non_positive_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_negative() or i->val.is_zero()));
}
void rat_negative_wrapper(expression_base::level l)
{ shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_negative()));
}

void rat_eq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}

void rat_neq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void rat_less_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<j->val));
}

void rat_lesseq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<=j->val));
}

void rat_greater_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>j->val));
}

void rat_greatereq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>=j->val));
}

void equiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(a==b));
}

void inequiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(a!=b));
}

void string_unary_eq_wrapper(expression_base::level l)
{ shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.empty()));
}
void string_unary_neq_wrapper(expression_base::level l)
{ shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.size()>0));
}
void string_eq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}
void string_neq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void string_less_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<j->val));
}
void string_leq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<=j->val));
}
void string_greater_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>j->val));
}
void string_geq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>=j->val));
}

void string_concatenate_wrapper(expression_base::level l)
{ shared_string b=get<string_value>(); shared_string a=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(a->val+b->val));
}

void concatenate_strings_wrapper(expression_base::level l)
{ shared_row arg=get<row_value>();
  if (l==expression_base::no_value)
    return;
  const std::vector<shared_value>& x=arg->val;
  std::vector<const std::string*> p; p.reserve(x.size());
  for (auto it=x.cbegin(); it!=x.cend(); ++it)
    p.push_back(&force<string_value>(it->get())->val);
  std::size_t s=0;
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    s+=(*it)->size();
  std::string result(s,char()); auto dst=result.begin();
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    dst=std::copy((*it)->begin(),(*it)->end(),dst);
  assert(dst==result.end());
  push_value(std::make_shared<string_value>(std::move(result)));
}

void string_to_ascii_wrapper(expression_base::level l)
{ shared_string c=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
      (c->val.size()==0 ? -1 : (unsigned char)c->val[0]));
}

void ascii_char_wrapper(expression_base::level l)
{ int c=get<int_value>()->int_val();
  if ((c<' ' and c!='\n') or c>'~')
    throw runtime_error() << "Value " << c << " out of printable ASCII range";
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(std::string(1,c)));
}

void matrix_shape_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<int_value>(m->val.numRows()));
  push_value(std::make_shared<int_value>(m->val.numColumns()));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void vector_suffix_wrapper(expression_base::level l)
{ int e=get<int_value>()->int_val();
  own_vector r=get_own<vector_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(e);
    push_value(r);
  }
}

void vector_prefix_wrapper(expression_base::level l)
{ own_vector r=get_own<vector_value>();
  int e=get<int_value>()->int_val();
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(r);
  }
}

void join_vectors_wrapper(expression_base::level l)
{ shared_vector y=get<vector_value>();
  shared_vector x=get<vector_value>();
  if (l!=expression_base::no_value)
  { own_vector result = std::make_shared<vector_value>(int_Vector());
    result->val.reserve(x->val.size()+y->val.size());
    result->val.insert(result->val.end(),x->val.begin(),x->val.end());
    result->val.insert(result->val.end(),y->val.begin(),y->val.end());
    push_value(std::move(result));
  }

}

void join_vector_row_wrapper(expression_base::level l)
{ shared_row arg=get<row_value>();
  if (l==expression_base::no_value)
    return;
  const std::vector<shared_value>& x=arg->val;
  std::vector<const int_Vector*> p; p.reserve(x.size());
  for (auto it=x.cbegin(); it!=x.cend(); ++it)
    p.push_back(&force<vector_value>(it->get())->val);
  std::size_t s=0;
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    s+=(*it)->size();
  int_Vector result(s);
  auto dst=result.begin();
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    dst=std::copy((*it)->cbegin(),(*it)->cend(),dst);
  assert(dst==result.end());
  push_value(std::make_shared<vector_value>(std::move(result)));
}

void vec_unary_eq_wrapper(expression_base::level l)
{ shared_vector i=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const auto end=i->val.end();
  for (auto it=i->val.begin(); it!=end; ++it)
    if (*it!=0)
    { push_value(whether(false));
      return; }
  push_value(whether(true));
}
void vec_unary_neq_wrapper(expression_base::level l)
{ shared_vector i=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const auto end=i->val.end();
  for (auto it=i->val.begin(); it!=end; ++it)
    if (*it!=0)
    { push_value(whether(true));
      return; }
  push_value(whether(false));
}
void vec_eq_wrapper(expression_base::level l)
{ shared_vector j=get<vector_value>(); shared_vector i=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}
void vec_neq_wrapper(expression_base::level l)
{ shared_vector j=get<vector_value>(); shared_vector i=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void vec_non_negative_wrapper(expression_base::level l)
{ shared_vector v = get<vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.begin(); it!=v->val.end(); ++it)
    if (*it<0)
    { OK=false; break; }
  push_value(whether(OK));
}
void vec_positive_wrapper(expression_base::level l)
{ shared_vector v = get<vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.begin(); it!=v->val.end(); ++it)
    if (*it<=0)
    { OK=false; break; }
  push_value(whether(OK));
}

void ratvec_unary_eq_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  push_value(whether(v->val.isZero()));
}
void ratvec_unary_neq_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  push_value(whether(not v->val.isZero()));
}
void ratvec_eq_wrapper(expression_base::level l)
{ shared_rational_vector w=get<rational_vector_value>();
  shared_rational_vector v=get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(v->val==w->val));
}
void ratvec_neq_wrapper(expression_base::level l)
{ shared_rational_vector w=get<rational_vector_value>();
  shared_rational_vector v=get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(v->val!=w->val));
}

void ratvec_non_negative_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.numerator().begin(); it!=v->val.numerator().end(); ++it)
    if (*it<0)
    { OK=false; break; }
  push_value(whether(OK));
}
void ratvec_positive_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.numerator().begin(); it!=v->val.numerator().end(); ++it)
    if (*it<=0)
    { OK=false; break; }
  push_value(whether(OK));
}

void mat_unary_eq_wrapper(expression_base::level l)
{ shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_zero()));
}
void mat_unary_neq_wrapper(expression_base::level l)
{ shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not i->val.is_zero()));
}
void mat_eq_wrapper(expression_base::level l)
{ shared_matrix j=get<matrix_value>(); shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}
void mat_neq_wrapper(expression_base::level l)
{ shared_matrix j=get<matrix_value>(); shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void check_size (std::size_t a, std::size_t b)
{ if (a!=b)
    throw runtime_error() << "Size mismatch " << a << ":" << b;
}

void vec_plus_wrapper(expression_base::level l)
{ own_vector v1= get_own<vector_value>();
  shared_vector v0= get<vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l==expression_base::no_value)
    return;
  v1->val += v0->val;
  push_value(std::move(v1));
}
void vec_minus_wrapper(expression_base::level l)
{ own_vector v1= get_own<vector_value>();
  shared_vector v0= get<vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l==expression_base::no_value)
    return;
  v1->val.negate_add(v0->val);
  push_value(std::move(v1));
}

void vec_unary_minus_wrapper(expression_base::level l)
{ own_vector v = get_own<vector_value>();
  if (l==expression_base::no_value)
    return;
  v->val.negate();
  push_value(std::move(v));
}

void vec_times_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_vector v= get_own<vector_value>();
  if (l==expression_base::no_value)
    return;
  v->val *= i;
  push_value(std::move(v));
}
void vec_divide_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_vector v= get_own<vector_value>();
  if (i==0)
    throw runtime_error("Vector division by 0");
  if (l==expression_base::no_value)
    return;
  divide(v->val,i);
  push_value(std::move(v));
}
void vec_modulo_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_vector v= get_own<vector_value>();
  if (i==0)
    throw runtime_error("Vector modulo 0");
  if (l==expression_base::no_value)
    return;
  v->val %= i;
  push_value(std::move(v));
}

void vv_prod_wrapper(expression_base::level l)
{ shared_vector w=get<vector_value>();
  shared_vector v=get<vector_value>();
  check_size (v->val.size(),w->val.size());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(v->val.dot(w->val)));
}

void flex_add_wrapper(expression_base::level l)
{ shared_vector v1= get<vector_value>();
  shared_vector v0= get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const int_Vector& V0=v0->val;
  auto i0=V0.size();
const int_Vector& V1=v1->val;
  auto i1=V1.size();
  while (i0>0 and V0[i0-1]==0)
    --i0;
  while (i1>0 and V1[i1-1]==0)
    --i1;
  if (i0==i1) // equal size case
  { while (i0>0 and V0[i0-1]+V1[i0-1]==0)
      --i0; // skip leading zeros in result
    int_Vector result(i0);
    while (i0-->0)
      result[i0]=V0[i0]+V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
    return;
  }
  shared_vector& larger = *(i0>i1 ? &v0 : &v1); // unequal size case
  if (larger.unique()) // then we can grab the larger vector
  { own_vector result =
      std::const_pointer_cast<vector_value>(std::move(larger));
    if (i0>i1)
    {
      result->val.resize(i0);
      while (i1-->0)
        result->val[i1]+=V1[i1]; // add |V1| to result
    }
    else
    {
      result->val.resize(i1);
      while (i0-->0)
        result->val[i0]+=V0[i0]; // add |V0| to result
    }
    push_value(std::move(result));
  }
  else // we must allocate
  {
    int_Vector result(std::max(i0,i1));
    while (i0>i1)
      --i0,result[i0]=V0[i0]; // copy excess of |V0|
    while (i1>i0)
      --i1,result[i1]=V1[i1]; // copy excess of |V1|
    while (i0-->0)
      result[i0]=V0[i0]+V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
  }
}

void flex_sub_wrapper(expression_base::level l)
{ shared_vector v1= get<vector_value>();
  shared_vector v0= get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const int_Vector& V0=v0->val;
  auto i0=V0.size();
const int_Vector& V1=v1->val;
  auto i1=V1.size();
  while (i0>0 and V0[i0-1]==0)
    --i0;
  while (i1>0 and V1[i1-1]==0)
    --i1;
  if (i0==i1) // equal size case
  { while (i0>0 and V0[i0-1]==V1[i0-1])
      --i0; // skip leading zeros in result
    int_Vector result(i0);
    while (i0-->0)
      result[i0]=V0[i0]-V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
    return;
  }
  shared_vector& larger = *(i0>i1 ? &v0 : &v1); // unequal size case
  if (larger.unique()) // then we can grab the larger vector
  { own_vector result =
      std::const_pointer_cast<vector_value>(std::move(larger));
    if (i0>i1)
    {
      result->val.resize(i0);
      while (i1-->0)
        result->val[i1]-=V1[i1]; // subtract |V1| from result
    }
    else
    {
      result->val.resize(i1);
      while (i1-->i0)
        result->val[i1]=-result->val[i1]; // negate top part of result
      while (i0-->0)
        result->val[i0]=V0[i0]-result->val[i0]; // negate result, add |V0|
    }
    push_value(std::move(result));
  }
  else // we must allocate
  {
    int_Vector result(std::max(i0,i1));
    while (i0>i1)
      --i0,result[i0]=V0[i0]; // copy excess of |V0|
    while (i1>i0)
      --i1,result[i1]=-V1[i1]; // copy excess of |V1|, negated
    while (i0-->0)
      result[i0]=V0[i0]-V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
  }
}

void vector_convolve_wrapper(expression_base::level l)
{ shared_vector v1= get<vector_value>();
  shared_vector v0= get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const int_Vector& V0=v0->val;
  auto i0=V0.size();
const int_Vector& V1=v1->val;
  auto i1=V1.size();
  while (i0>0 and V0[i0-1]==0)
    --i0;
  while (i1>0 and V1[i1-1]==0)
    --i1;
  if (i0==0 or i1==0)
{ push_value(std::make_shared<vector_value>(int_Vector(0)));
    return;
  }
  own_vector result = std::make_shared<vector_value>(int_Vector(i0+i1-1));
  int_Vector& r = result->val;
  auto i=i0; std::size_t j=0; int V1j=V1[0], V0l=V0[i0-1];
  while (i-->0)
    r[i] = V0[i]*V1j; // copy |V0|, multiplied by lowest (constant) term of |V1|
  while (++j<i1)
  { r[(i=i0-1)+j] = V0l*(V1j=V1[j]); // copy top term of |V0| times next of |V1|
    while (i-->0)
      r[i+j] += V0[i]*V1j; // add remainder of multiple of |V0|, shifted |j|
  }
  push_value(std::move(result));
}

void vector_div_wrapper(expression_base::level l)
{ int n=get<int_value>()->int_val();
  own_vector v=get_own<vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>
      (std::move(v->val),n)); // throws if |n==0|
}

void ratvec_unfraction_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l!=expression_base::no_value)
  { Weight num(v->val.numerator().begin(),v->val.numerator().end()); // convert
    push_value(std::make_shared<vector_value>(std::move(num)));
    push_value(std::make_shared<int_value>(v->val.denominator()));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void ratvec_plus_wrapper(expression_base::level l)
{ shared_rational_vector v1= get<rational_vector_value>();
  shared_rational_vector v0= get<rational_vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v0->val+v1->val));
}
void ratvec_minus_wrapper(expression_base::level l)
{ shared_rational_vector v1= get<rational_vector_value>();
  shared_rational_vector v0= get<rational_vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v0->val-v1->val));
}
void ratvec_unary_minus_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(-v->val));
}

void ratvec_times_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_rational_vector v= get_own<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  (v->val *= i).normalize();
  push_value(v);
}
void ratvec_divide_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_rational_vector v= get_own<rational_vector_value>();
  if (i==0)
    throw runtime_error("Rational vector division by 0");
  if (l==expression_base::no_value)
    return;
  (v->val /= i).normalize();
  push_value(v);
}
void ratvec_modulo_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_rational_vector v= get_own<rational_vector_value>();
  if (i==0)
    throw runtime_error("Rational vector modulo 0");
  if (l==expression_base::no_value)
    return;
  v->val %= i;
  push_value(v);
}


void ratvec_times_rat_wrapper(expression_base::level l)
{ shared_rat r= get<rat_value>();
  own_rational_vector v= get_own<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  (v->val *= r->rat_val()).normalize();
  push_value(v);
}
void ratvec_divide_rat_wrapper(expression_base::level l)
{ shared_rat r= get<rat_value>();
  own_rational_vector v= get_own<rational_vector_value>();
  if (r->val.is_zero())
    throw runtime_error("Rational vector division by 0");
  if (l==expression_base::no_value)
    return;
  (v->val /= r->rat_val()).normalize();
  push_value(v);
}

void mat_plus_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_matrix m= get_own<matrix_value>();
  if (l==expression_base::no_value)
    return;
  m->val += i;
  push_value(m);
}
void mat_minus_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->int_val();
  own_matrix m= get_own<matrix_value>();
  if (l==expression_base::no_value)
    return;
  m->val += -i;
  push_value(m);
}

void int_plus_mat_wrapper(expression_base::level l)
{ own_matrix m= get_own<matrix_value>();
  int i= get<int_value>()->int_val();
  if (l==expression_base::no_value)
    return;
  m->val += i;
  push_value(m);
}
void int_minus_mat_wrapper(expression_base::level l)
{ own_matrix m= get_own<matrix_value>();
  int i= get<int_value>()->int_val();
  if (l==expression_base::no_value)
    return;
  m->val.negate();
  m->val += i;
  push_value(m);
}

void mat_plus_mat_wrapper(expression_base::level l)
{ own_matrix b= get_own<matrix_value>();
  shared_matrix a= get<matrix_value>();
  check_size(a->val.numRows(),b->val.numRows());
  check_size(a->val.numColumns(),b->val.numColumns());
  if (l==expression_base::no_value)
    return;
  b->val += a->val;
  push_value(b);
}

void mat_minus_mat_wrapper(expression_base::level l)
{
  shared_matrix b= get<matrix_value>();
  own_matrix a= get_own<matrix_value>();
  check_size(a->val.numRows(),b->val.numRows());
  check_size(a->val.numColumns(),b->val.numColumns());
  if (l==expression_base::no_value)
    return;
  a->val -= b->val;
  push_value(a);
}

void mv_prod_wrapper(expression_base::level l)
{ shared_vector v=get<vector_value>();
  shared_matrix m=get<matrix_value>();
  if (m->val.numColumns()!=v->val.size())
    throw runtime_error() << "Size mismatch " 
      << m->val.numColumns() << ':' << v->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(m->val*v->val));
}

void mrv_prod_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  shared_matrix m=get<matrix_value>();
  if (m->val.numColumns()!=v->val.size())
    throw runtime_error() << "Size mismatch " 
      << m->val.numColumns() << ':' << v->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(m->val*v->val));
}

void vm_prod_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>(); // right factor
  shared_vector v=get<vector_value>(); // left factor
  if (v->val.size()!=m->val.numRows())
    throw runtime_error()
      << "Size mismatch " << v->val.size() << ":" << m->val.numRows();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(m->val.right_prod(v->val)));
}

void rvm_prod_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  shared_rational_vector v=get<rational_vector_value>();
  if (v->val.size()!=m->val.numRows())
    throw runtime_error() << "Size mismatch " 
      << v->val.size() << ':' << m->val.numRows();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v->val*m->val));
}

void null_vec_wrapper(expression_base::level l)
{ int n=get<int_value>()->int_val();
  if (n<0)
    throw runtime_error() << "Negative size for vector: " << n;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(int_Vector(n,0)));
}
 void null_mat_wrapper(expression_base::level l)
{ int n=get<int_value>()->int_val();
  int m=get<int_value>()->int_val();
  if (m<0)
    throw runtime_error() << "Negative number of rows: " << m;
  if (n<0)
    throw runtime_error() << "Negative number of columns: " << n;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value> (int_Matrix(m,n,0)));
}
void transpose_vec_wrapper(expression_base::level l)
{ shared_vector v=get<vector_value>();
  if (l!=expression_base::no_value)
  { own_matrix m = std::make_shared<matrix_value>(int_Matrix(1,v->val.size()));
    for (std::size_t j=0; j<v->val.size(); ++j)
      m->val(0,j)=v->val[j];
    push_value(std::move(m));
  }
}

void diagonal_wrapper(expression_base::level l)
{ shared_vector d=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  auto n=d->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,n,0));
  for (std::size_t i=0; i<n; ++i)
    m->val(i,i)=d->val[i];
  push_value(std::move(m));
}

void stack_rows_wrapper(expression_base::level l)
{ shared_row r = get<row_value>();
  auto n = r->val.size();
  std::vector<const int_Vector*> row(n);
  std::size_t width=0; // maximal length of vectors
  for(std::size_t i=0; i<n; ++i)
  { row[i] = & force<vector_value>(r->val[i].get())->val;
    if (row[i]->size()>width)
      width=row[i]->size();
  }

  if (l==expression_base::no_value)
    return;

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,width,0));
  for(std::size_t i=0; i<n; ++i)
    for (std::size_t j=0; j<row[i]->size(); ++j)
      m->val(i,j)=(*row[i])[j];
  push_value(std::move(m));
}

void combine_columns_wrapper(expression_base::level l)
{ shared_row r = get<row_value>();
  int n = get<int_value>()->int_val();
  if (n<0)
    throw runtime_error() << "Negative number " << n <<" of rows requested";

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(std::size_t j=0; j<r->val.size(); ++j)
  { const int_Vector& col = force<vector_value>(r->val[j].get())->val;
    if (col.size()!=std::size_t(n))
      throw runtime_error() << "Column " << j <<" size " << col.size() 
         << " does not match specified size " << n;

    m->val.set_column(j,col);
  }
  if (l!=expression_base::no_value)
    push_value(std::move(m));
}

void combine_rows_wrapper(expression_base::level l)
{ shared_row r =get<row_value>();
  int n = get<int_value>()->int_val();
  if (n<0)
    throw runtime_error() << "Negative number " << n << " of columns requested";

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(r->val.size(),n));
  for(std::size_t i=0; i<r->val.size(); ++i)
  { const int_Vector& row = force<vector_value>(r->val[i].get())->val;
    if (row.size()!=std::size_t(n))
      throw runtime_error() << "Row " << i << " size " << row.size() 
        << " does not match specified size " << n;

    m->val.set_row(i,row);
  }
  if (l!=expression_base::no_value)
    push_value(std::move(m));
}

template <bool transpose, bool negate>
void transform_copy
  (unsigned flags,
   const int_Matrix& src, 
   int lwb_r, int upb_r,
   int lwb_c, int upb_c,
   int_Matrix& dst)
{ if ((flags&0x1)==0) // no reversal of rows
  { if ((flags&0x2)==0) // no reversal of rows or columns
      for (int i=0, k=lwb_r; k<upb_r; ++i, ++k)
	for (int j=0, l=lwb_c; l<upb_c; ++j, ++l)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
    else // reversal of columns only
      for (int i=0, k=lwb_r; k<upb_r; ++i, ++k)
	for (int j=0, l=upb_c; l-->lwb_c; ++j)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
  }
  else // reversal of rows
  { if ((flags&0x2)==0) // reversal of rows only
      for (int i=0, k=upb_r; k-->lwb_r; ++i)
	for (int j=0, l=lwb_c; l<upb_c; ++j, ++l)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
    else // reversal of rows and columns
      for (int i=0, k=upb_r; k-->lwb_r; ++i)
	for (int j=0, l=upb_c; l-->lwb_c; ++j)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
  }
}

void swiss_matrix_knife_wrapper(expression_base::level lev)
{ int l = get<int_value>()->int_val();
  int j = get<int_value>()->int_val();
  int k = get<int_value>()->int_val();
  int i = get<int_value>()->int_val();
  shared_matrix src = get<matrix_value>();
  const int_Matrix& A = src->val;
  BitSet<8> flags (get<int_value>()->int_val());

  int m = A.numRows(); int n= A.numColumns();
  int lwb_r = flags[1] ? m-i : i;
  int upb_r = flags[2] ? m-k : k;
  int lwb_c = flags[4] ? n-j : j;
  int upb_c = flags[5] ? n-l : l;
  if (lwb_r<0 or upb_r>m or lwb_c<0 or upb_c>n)
    
    { runtime_error e;
      e << "Range exceeds bounds: ";
      if (lwb_r<0 or upb_r>m)
        if (upb_r>m)
          if (lwb_r<0)
            e << "both row bounds " << lwb_r << ':' << upb_r;
          else
            e << "upper row bound " << upb_r;
        else
          e << "lower row bound " << lwb_r;
      if ((lwb_r<0 or upb_r>m) and (lwb_c<0 or upb_c>n))
        e << " and ";
      if (lwb_c<0 or upb_c>n)
        if (upb_c>n)
          if (lwb_c<0)
            e << "both column bounds " << lwb_c << ':' << upb_c;
          else
            e << "upper column bound " << upb_c;
        else
          e << "lower column bound " << lwb_c;
      e << " out of range, actual bounds 0:" << m << ", 0:" << n;
      throw e;
    }

  if (lev==expression_base::no_value)
    return;
  
  if (lwb_r>upb_r)
    upb_r = lwb_r;
  if (lwb_c>upb_c)
    upb_c = lwb_c;
  unsigned r_size = upb_r - lwb_r;
  unsigned c_size = upb_c - lwb_c;
  if (flags[6]) // transpose
    std::swap(r_size,c_size);
  own_matrix result(std::make_shared<matrix_value>(int_Matrix(r_size,c_size)));
  unsigned rev_flags = static_cast<unsigned>(flags[0])*0x1
                     ^ static_cast<unsigned>(flags[3])*0x2;
  
  if (flags[6])
  { if (flags[7])
      transform_copy<true,true>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
    else
      transform_copy<true,false>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
  }
  else
  { if (flags[7])
      transform_copy<false,true>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
    else
      transform_copy<false,false>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
  }
  push_value(std::move(result));
}

void gcd_wrapper(expression_base::level l)
{ own_vector v=get_own<vector_value>();
  if (l==expression_base::no_value)
    return;
  bool flip=false;
  int d =
    matreduc::gcd(std::move(v->val),static_cast<int_Matrix*>(nullptr),flip);
  push_value(std::make_shared<int_value>(d));
}

void Bezout_wrapper(expression_base::level l)
{ own_vector v=get_own<vector_value>();
  if (l==expression_base::no_value)
    return;
  own_matrix column = std::make_shared<matrix_value>(int_Matrix());
  bool flip=false;
  int d = matreduc::gcd(std::move(v->val),&column->val,flip);
  push_value(std::make_shared<int_value>(d));
  push_value(std::move(column));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void echelon_wrapper(expression_base::level l)
{ own_matrix A=get_own<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_matrix column = std::make_shared<matrix_value>(int_Matrix());
  bool flip;
  BitMap pivots=matreduc::column_echelon(A->val,column->val,flip);
  push_value(std::move(A));
  push_value(std::move(column));
  own_row p_list = std::make_shared<row_value>(0);
  p_list->val.reserve(pivots.size());
  for (BitMap::iterator it=pivots.begin(); it(); ++it)
    p_list->val.push_back(std::make_shared<int_value>(*it));
  push_value(std::move(p_list));
  push_value(std::make_shared<int_value>(flip ? -1 : 1));
  if (l==expression_base::single_value)
    wrap_tuple<4>();
}

void linear_solve_wrapper(expression_base::level l)
{ own_vector b=get_own<vector_value>();
  own_matrix A=get_own<matrix_value>();
  if (A->val.numRows()!=b->val.size())
    throw runtime_error() << "Linear system size mismatch "
                          << A->val.numRows() << ':' << b->val.size();
  if (l==expression_base::no_value)
    return;
  own_matrix column = std::make_shared<matrix_value>(int_Matrix());
  bool flip; // unused
  const auto m = A->val.numColumns();
  BitMap pivots=matreduc::column_echelon(A->val,column->val,flip);
  try
  { static id_type affine_name=main_hash_table->match_literal("affine_subspace");
    const auto k = A->val.numColumns(); // number of pivots
    arithmetic::big_int factor;
    int_Vector ini_sol = matreduc::echelon_solve(A->val,pivots,b->val,factor);
    push_value(std::make_shared<vector_value> 
      (column->val.block(0,0,m,k)*ini_sol));
    push_value(std::make_shared<int_value>(std::move(factor)));
    push_value(std::make_shared<matrix_value>(column->val.block(0,k,m,m)));
    wrap_tuple<3>();
    push_value(std::make_shared<union_value>(1,pop_value(),affine_name));
  }
  catch(const std::runtime_error& e)
  { static id_type empty_name=main_hash_table->match_literal("empty_set");
    auto empty = std::make_shared<tuple_value>(0);
    push_value(std::make_shared<union_value>(0,std::move(empty),empty_name));
  }
}

void diagonalize_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
  { own_matrix row = std::make_shared<matrix_value>(int_Matrix());
    own_matrix column = std::make_shared<matrix_value>(int_Matrix());
    own_vector diagonal = std::make_shared<vector_value>
       (matreduc::diagonalise(M->val,row->val,column->val));
    push_value(std::move(diagonal));
    push_value(std::move(row));
    push_value(std::move(column));
    if (l==expression_base::single_value)
      wrap_tuple<3>();
  }
}

void adapted_basis_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
  { own_vector diagonal = std::make_shared<vector_value>(std::vector<int>());
    push_value(std::make_shared<matrix_value>
      (matreduc::adapted_basis(M->val,diagonal->val)));
    push_value(std::move(diagonal));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void kernel_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lattice::kernel(M->val)));
}

void eigen_lattice_wrapper(expression_base::level l)
{ int eigen_value = get<int_value>()->int_val();
  shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (lattice::eigen_lattice(M->val,eigen_value)));
}

void row_saturate_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lattice::row_saturate(M->val)));
}

void Smith_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
push_value(std::make_shared<matrix_value>
    (matreduc::Smith_basis(m->val,inv_factors->val)));
  push_value(std::move(inv_factors));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void invert_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (m->val.numRows()!=m->val.numColumns())
    throw runtime_error()
      << "Cannot invert a " 
      << m->val.numRows() << "x" << m->val.numColumns() << " matrix";
  if (l==expression_base::no_value)
    return;
  arithmetic::big_int denom;
push_value(std::make_shared<matrix_value>(inverse(m->val,denom)));
  push_value(std::make_shared<int_value>(std::move(denom)));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void section_wrapper(expression_base::level l)
{
  shared_matrix m=get<matrix_value>();
  BinaryMap A(m->val);
  BinaryMap B=A.section();
  own_matrix res = std::make_shared<matrix_value>(
    int_Matrix(B.numRows(),B.numColumns()));
  for (unsigned int j=B.numColumns(); j-->0;)
    res->val.set_column(j,int_Vector(B.column(j)));
  if (l!=expression_base::no_value)
    push_value(std::move(res));
}

void subspace_normal_wrapper(expression_base::level l)
{
  typedef BitVector<64> bitvec;
  shared_matrix generators=get<matrix_value>();
  unsigned int n_gens = generators->val.numColumns();
  unsigned int dim = generators->val.numRows();
  if (dim>64)
    throw runtime_error() << "Dimension too large: " << dim << ">64";
  if (n_gens>64)
    throw runtime_error() << "Too many generators: " << n_gens << ">64";

  std::vector<bitvec> basis, combination;
    // |basis[j]| will be initialised from column $j$ of |generators|
  std::vector<unsigned int> pivot; // |pivot[i]| is bit position for |basis[i]|
  std::vector<unsigned int> pivoter;
    // generator |pivoter[i]| led to |basis[i]|
  { unsigned max_rank=std::min(n_gens,dim); // |basis| cannot exceed this size
    basis.reserve(max_rank); pivot.reserve(max_rank); pivoter.reserve(max_rank);
  }
  bitvector::initBasis(combination,n_gens);
    // express (still virtual) |basis| elements in |generators|

  
  for (unsigned int j=0; j<n_gens; ++j)
  { bitvec v(generators->val.column(j)); // reduce modulo $2$
    for (unsigned int l=0; l<basis.size(); ++l)
      if (v[pivot[l]])
      {
         v -= basis[l];
         combination[j] -= combination[pivoter[l]];
      }
    if (v.nonZero())
    {
      unsigned int piv = v.firstBit(); // new pivot
      for (unsigned int l=0; l<basis.size(); ++l)
        if (basis[l][piv])
        {
           basis[l] -= v;
           combination[pivoter[l]] -= combination[j];
        }
      basis.push_back(v);
      pivoter.push_back(j);
      pivot.push_back(piv);
    }
  }
  if (l==expression_base::no_value)
    return;

  
  { Permutation pi = permutations::standardization(pivot,dim);
      // relative positions of pivots
    unsigned rank=basis.size(); // dimension of the subspace
    int_Matrix basis_m(dim,rank,0);
    int_Matrix combin_m(n_gens,rank,0);
    int_Matrix relations_m(n_gens,n_gens-rank);
    own_row pivot_r = std::make_shared<row_value>(rank);
    unsigned int l=0; // number of basis vectors copied so far, current index
    for (unsigned int j=0; j<n_gens; ++j)
      if (l<rank and j==pivoter[l])
      { unsigned d = pi[l]; // destination position
        for (auto it=basis[l].data().begin(); it(); ++it)
          basis_m(*it,d) = 1;
        for (auto it= combination[j].data().begin(); it(); ++it)
          combin_m(*it,d) = 1;
        pivot_r->val[d] = std::make_shared<int_value>(pivot[l]);
        ++l;
      }
      else
      { unsigned d = j-l;
        for (auto it= combination[j].data().begin(); it(); ++it)
          relations_m(*it,d) = 1;
      }
    assert (l==rank);
  push_value(std::make_shared<matrix_value>(std::move(basis_m)));
    push_value(std::make_shared<matrix_value>(std::move(combin_m)));
    push_value(std::make_shared<matrix_value>(std::move(relations_m)));
    push_value(std::move(pivot_r));
  }
  if (l==expression_base::single_value)
    wrap_tuple<4>();
}

}

void Id_table::add(id_type id, shared_value val, type_expr&& type, bool is_const)
{ auto its = table.equal_range(id);

  if (its.first==its.second) // no global identifier was previously known
    table.emplace_hint(its.first,id, id_data 
    (std::make_shared<shared_value>(std::move(val)),std::move(type),is_const));
  else // a global identifier was previously known
    its.first->second = id_data(
      std::make_shared<shared_value>(std::move(val)), std::move(type),is_const);
}

void Id_table::add_type_def(id_type id, type_expr&& type)
{ auto its = table.equal_range(id);

  if (its.first==its.second) // no global identifier was previously known
    table.emplace_hint 
      (its.first,id,id_data(shared_share(),std::move(type),true));
  else // a global identifier was previously known, replace it
    its.first->second = id_data(shared_share(),std::move(type),true);
}

bool Id_table::is_defined_type(id_type id) const
{ map_type::const_iterator p = table.find(id);
 return p!=table.end() and p->second.value()==nullptr;
}

bool Id_table::remove(id_type id)
{ map_type::iterator p = table.find(id);
  if (p==table.end())
    return false;
  table.erase(p); return true;
}

const_type_p Id_table::type_of(id_type id,bool& is_const) const
{ map_type::const_iterator p=table.find(id);
  if (p==table.end())
  { is_const=false; return nullptr; } // avoid later ``uninintialized'' warning
  is_const=p->second.is_const();
  return &p->second.type();
}
const_type_p Id_table::type_of(id_type id) const
{ map_type::const_iterator p=table.find(id);
  if (p==table.end())
    return nullptr;
  return &p->second.type();
}

void Id_table::specialise(id_type id,const type_expr& type)
{ map_type::iterator p=table.find(id);
  p->second.type().specialise(type);
}

shared_value Id_table::value_of(id_type id) const
{ map_type::const_iterator p=table.find(id);
  return p==table.end() ? shared_value(value(nullptr)) : *p->second.value();
}
shared_share Id_table::address_of(id_type id)
{ map_type::iterator p=table.find(id);
  if (p==table.end())
    throw logic_error() << "Identifier without table entry:" 
       << main_hash_table->name_of(id);

  return p->second.value();
}

void Id_table::print(std::ostream& out) const
{ for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
  { out << main_hash_table->name_of(p->first);
    const shared_share& v= p->second.value();
    if (v==nullptr)
      out << " = " << p->second.type();
    else
    { out << ": " << p->second.type() << ": ";
      if (*v==nullptr)
        out << '*';
      else
        out << **p->second.value();
     }
     out << std::endl;
   }
}

std::ostream& operator<< (std::ostream& out, const Id_table& p)
{ p.print(out); return out; }

const overload_table::variant_list& overload_table::variants
  (id_type id) const
{ static const variant_list empty;
  auto p=table.find(id);
  return p==table.end() ? empty : p->second;
}
const overload_data* overload_table::entry
  (id_type id, const type_expr& arg_t) const
{ const variant_list& vars = variants(id);
  for (auto it=vars.begin(); it!=vars.end(); ++it)
    if (it->type().arg_type==arg_t)
      return &*it;
  return nullptr; // indicate not found
}

void overload_table::add
  (id_type id, shared_function val, type_expr&& t)
{ assert (t.kind()==function_type);
  func_type ftype(t.func()->copy()); // locally copy the function type
  auto its = table.equal_range(id);
  if (its.first==its.second) // a fresh overloaded identifier
  {
    auto pos=table.emplace_hint(its.first,id,variant_list());
    pos->second.emplace_back(std::move(val), std::move(ftype) );
  }
  else
    
    { variant_list& slot=its.first->second; // vector of all variants
      auto pos=locate_overload(id,slot,ftype.arg_type); // may |throw|
      if (pos<slot.size() and slot[pos].type().arg_type==ftype.arg_type)
         // equality found
        slot[pos] = overload_data(std::move(val),std::move(ftype)); // overwrite
      else
        slot.emplace(slot.begin()+pos,std::move(val),std::move(ftype));
    }
}

bool overload_table::remove(id_type id, const type_expr& arg_t)
{ map_type::iterator p=table.find(id);
  if (p==table.end()) return false; // |id| was not known at all
  variant_list& variants=p->second;
  for (std::size_t i=0; i<variants.size(); ++i)
    if (variants[i].type().arg_type==arg_t)
    {
      variants.erase(variants.begin()+i);
      return true;
    }
  return false; // |id| was known, but no such overload is present
}

void overload_table::print(std::ostream& out) const
{ for (auto p=table.begin(); p!=table.end(); ++p)
    for (auto it=p->second.begin(); it!=p->second.end(); ++it)
      out << main_hash_table->name_of(p->first) << ": " 
        << it->type() << ": " << *it->value() << std::endl;
}

std::ostream& operator<< (std::ostream& out, const overload_table& p)
{ p.print(out); return out; }

void initialise_evaluator()
{ 
  execution_stack.reserve(16); 
  
  coercion(int_type,rat_type, "QI", rational_convert); 
  coercion(row_of_int_type, vec_type, "V[I]", intlist_vector_convert); 
  coercion(row_of_rat_type,ratvec_type, "Qv[Q]", ratlist_ratvec_convert); 
  coercion(vec_type,row_of_int_type, "[I]V", vector_intlist_convert); 
  coercion(ratvec_type,row_of_rat_type, "[Q]Qv", ratvec_ratlist_convert); 
  coercion(vec_type,ratvec_type,"QvV", vector_ratvec_convert); 
  coercion(row_of_int_type,ratvec_type,"Qv[I]", intlist_ratvec_convert); 
  coercion(vec_type,row_of_rat_type,"[Q]V", vector_ratlist_convert); 
  coercion(row_of_int_type,row_of_rat_type,"[Q][I]", intlist_ratlist_convert);
  
  coercion(row_of_vec_type,mat_type, "M[V]", veclist_matrix_convert); 
  coercion(row_row_of_int_type,mat_type, "M[[I]]", intlistlist_matrix_convert); 
  coercion(row_row_of_int_type,row_of_vec_type, "[V][[I]]",
    intlistlist_veclist_convert); 
  coercion(mat_type,row_of_vec_type, "[V]M", matrix_veclist_convert); 
  coercion(mat_type,row_row_of_int_type, "[[I]]M", matrix_intlistlist_convert); 
  coercion(row_of_vec_type,row_row_of_int_type, "[[I]][V]",
    veclist_intlistlist_convert); 
  coercion(mat_type,row_of_ratvec_type, "[Qv]M", matrix_ratveclist_convert); 
  coercion(mat_type,row_row_of_rat_type, "[[Q]]M", matrix_ratlistlist_convert); 
  coercion(row_of_vec_type,row_of_ratvec_type, "[Qv][V]",
     veclist_ratveclist_convert); 
  coercion(row_of_vec_type,row_row_of_rat_type, "[[Q]][V]",
     veclist_ratlistlist_convert); 
  coercion(row_row_of_int_type,row_of_ratvec_type, "[Qv][[I]]",
     intlistlist_ratveclist_convert); 
  coercion(row_row_of_int_type,row_row_of_rat_type, "[[Q]][[I]]",
      intlistlist_ratlistlist_convert);
  
  install_function(plus_wrapper,"+","(int,int->int)");
  install_function(minus_wrapper,"-","(int,int->int)");
  install_function(times_wrapper,"*","(int,int->int)");
  install_function(divide_wrapper,"\\","(int,int->int)");
  install_function(modulo_wrapper,"%","(int,int->int)");
  install_function(divmod_wrapper,"\\%","(int,int->int,int)");
  install_function(unary_minus_wrapper,"-","(int->int)");
  install_function(power_wrapper,"^","(int,int->int)");
  install_function(fraction_wrapper,"/","(int,int->rat)");
  install_function(unfraction_wrapper,"%","(rat->int,int)");
     // unary \% means ``break open''
  install_function(rat_plus_int_wrapper,"+","(rat,int->rat)");
  install_function(rat_minus_int_wrapper,"-","(rat,int->rat)");
  install_function(rat_times_int_wrapper,"*","(rat,int->rat)");
  install_function(rat_divide_int_wrapper,"/","(rat,int->rat)");
  install_function(rat_modulo_int_wrapper,"%","(rat,int->rat)");
  install_function(rat_quotient_int_wrapper,"\\","(rat,int->int)");
  install_function(rat_plus_wrapper,"+","(rat,rat->rat)");
  install_function(rat_minus_wrapper,"-","(rat,rat->rat)");
  install_function(rat_times_wrapper,"*","(rat,rat->rat)");
  install_function(rat_divide_wrapper,"/","(rat,rat->rat)");
  install_function(rat_modulo_wrapper,"%","(rat,rat->rat)");
  install_function(rat_unary_minus_wrapper,"-","(rat->rat)");
  install_function(rat_inverse_wrapper,"/","(rat->rat)");
  install_function(rat_floor_wrapper,"floor","(rat->int)");
  install_function(rat_ceil_wrapper,"ceil","(rat->int)");
  install_function(rat_frac_wrapper,"frac","(rat->rat)");
  install_function(rat_power_wrapper,"^","(rat,int->rat)");
  install_function(int_unary_eq_wrapper,"=","(int->bool)");
  install_function(int_unary_neq_wrapper,"!=","(int->bool)");
  install_function(int_non_negative_wrapper,">=","(int->bool)");
  install_function(int_positive_wrapper,">","(int->bool)");
  install_function(int_non_positive_wrapper,"<=","(int->bool)");
  install_function(int_negative_wrapper,"<","(int->bool)");
  install_function(int_eq_wrapper,"=","(int,int->bool)");
  install_function(int_neq_wrapper,"!=","(int,int->bool)");
  install_function(int_less_wrapper,"<","(int,int->bool)");
  install_function(int_lesseq_wrapper,"<=","(int,int->bool)");
  install_function(int_greater_wrapper,">","(int,int->bool)");
  install_function(int_greatereq_wrapper,">=","(int,int->bool)");
  install_function(rat_unary_eq_wrapper,"=","(rat->bool)");
  install_function(rat_unary_neq_wrapper,"!=","(rat->bool)");
  install_function(rat_non_negative_wrapper,">=","(rat->bool)");
  install_function(rat_positive_wrapper,">","(rat->bool)");
  install_function(rat_non_positive_wrapper,"<=","(rat->bool)");
  install_function(rat_negative_wrapper,"<","(rat->bool)");
  install_function(rat_eq_wrapper,"=","(rat,rat->bool)");
  install_function(rat_neq_wrapper,"!=","(rat,rat->bool)");
  install_function(rat_less_wrapper,"<","(rat,rat->bool)");
  install_function(rat_lesseq_wrapper,"<=","(rat,rat->bool)");
  install_function(rat_greater_wrapper,">","(rat,rat->bool)");
  install_function(rat_greatereq_wrapper,">=","(rat,rat->bool)");
  install_function(equiv_wrapper,"=","(bool,bool->bool)");
  install_function(inequiv_wrapper,"!=","(bool,bool->bool)");
  install_function(string_unary_eq_wrapper,"=","(string->bool)");
  install_function(string_unary_neq_wrapper,"!=","(string->bool)");
  install_function(string_eq_wrapper,"=","(string,string->bool)");
  install_function(string_neq_wrapper,"!=","(string,string->bool)");
  install_function(string_less_wrapper,"<","(string,string->bool)");
  install_function(string_leq_wrapper,"<=","(string,string->bool)");
  install_function(string_greater_wrapper,">","(string,string->bool)");
  install_function(string_geq_wrapper,">=","(string,string->bool)");
  install_function(string_concatenate_wrapper,"##","(string,string->string)");
  install_function(concatenate_strings_wrapper,"##","([string]->string)");
  install_function(string_to_ascii_wrapper,"ascii","(string->int)");
  install_function(ascii_char_wrapper,"ascii","(int->string)");
  install_function(sizeof_string_wrapper,"#","(string->int)");
  install_function(sizeof_vector_wrapper,"#","(vec->int)");
  install_function(sizeof_ratvec_wrapper,"#","(ratvec->int)");
  install_function(matrix_ncols_wrapper,"#","(mat->int)");
  install_function(vector_suffix_wrapper,"#","(vec,int->vec)");
  install_function(vector_prefix_wrapper,"#","(int,vec->vec)");
  install_function(join_vectors_wrapper,"##","(vec,vec->vec)");
  install_function(join_vector_row_wrapper,"##","([vec]->vec)");
  install_function(matrix_shape_wrapper,"shape","(mat->int,int)");
  install_function(vec_unary_eq_wrapper,"=","(vec->bool)");
  install_function(vec_unary_neq_wrapper,"!=","(vec->bool)");
  install_function(vec_eq_wrapper,"=","(vec,vec->bool)");
  install_function(vec_neq_wrapper,"!=","(vec,vec->bool)");
  install_function(vec_non_negative_wrapper,">=","(vec->bool)");
  install_function(vec_positive_wrapper,">","(vec->bool)");
  install_function(ratvec_unary_eq_wrapper,"=","(ratvec->bool)");
  install_function(ratvec_unary_neq_wrapper,"!=","(ratvec->bool)");
  install_function(ratvec_non_negative_wrapper,">=","(ratvec->bool)");
  install_function(ratvec_positive_wrapper,">","(ratvec->bool)");
  install_function(ratvec_eq_wrapper,"=","(ratvec,ratvec->bool)");
  install_function(ratvec_neq_wrapper,"!=","(ratvec,ratvec->bool)");
  install_function(mat_unary_eq_wrapper,"=","(mat->bool)");
  install_function(mat_unary_neq_wrapper,"!=","(mat->bool)");
  install_function(mat_eq_wrapper,"=","(mat,mat->bool)");
  install_function(mat_neq_wrapper,"!=","(mat,mat->bool)");
  install_function(vec_plus_wrapper,"+","(vec,vec->vec)");
  install_function(vec_minus_wrapper,"-","(vec,vec->vec)");
  install_function(vec_unary_minus_wrapper,"-","(vec->vec)");
  install_function(vec_times_int_wrapper,"*","(vec,int->vec)");
  install_function(vec_divide_int_wrapper,"\\","(vec,int->vec)");
  install_function(vec_modulo_int_wrapper,"%","(vec,int->vec)");
  install_function(vector_div_wrapper,"/","(vec,int->ratvec)");
  install_function(ratvec_unfraction_wrapper,"%","(ratvec->vec,int)");
  install_function(ratvec_plus_wrapper,"+","(ratvec,ratvec->ratvec)");
  install_function(ratvec_minus_wrapper,"-","(ratvec,ratvec->ratvec)");
  install_function(ratvec_unary_minus_wrapper,"-","(ratvec->ratvec)");
  install_function(ratvec_times_int_wrapper,"*","(ratvec,int->ratvec)");
  install_function(ratvec_divide_int_wrapper,"/","(ratvec,int->ratvec)");
  install_function(ratvec_modulo_int_wrapper,"%","(ratvec,int->ratvec)");
  install_function(ratvec_times_rat_wrapper,"*","(ratvec,rat->ratvec)");
  install_function(ratvec_divide_rat_wrapper,"/","(ratvec,rat->ratvec)");
  install_function(mat_plus_int_wrapper,"+","(mat,int->mat)");
  install_function(mat_minus_int_wrapper,"-","(mat,int->mat)");
  install_function(int_plus_mat_wrapper,"+","(int,mat->mat)");
  install_function(int_minus_mat_wrapper,"-","(int,mat->mat)");
  install_function(vv_prod_wrapper,"*","(vec,vec->int)");
  install_function(flex_add_wrapper,"flex_add","(vec,vec->vec)");
  install_function(flex_sub_wrapper,"flex_sub","(vec,vec->vec)");
  install_function(vector_convolve_wrapper,"convolve","(vec,vec->vec)");
  install_function(mat_plus_mat_wrapper,"+","(mat,mat->mat)");
  install_function(mat_minus_mat_wrapper,"-","(mat,mat->mat)");
  install_function(mrv_prod_wrapper,"*","(mat,ratvec->ratvec)");
  install_function(mv_prod_wrapper,"*","(mat,vec->vec)");
  install_function(mm_prod_wrapper,"*","(mat,mat->mat)");
  install_function(vm_prod_wrapper,"*","(vec,mat->vec)");
  install_function(rvm_prod_wrapper,"*","(ratvec,mat->ratvec)");
  
  install_function(null_vec_wrapper,"null","(int->vec)");
  install_function(null_mat_wrapper,"null","(int,int->mat)");
  install_function(transpose_vec_wrapper,"^","(vec->mat)");
  install_function(transpose_mat_wrapper,"^","(mat->mat)");
    // install as operator
  install_function(transpose_mat_wrapper,"transpose ","(mat->mat)");
    // use of space in the name makes this copy untouchable
  install_function(id_mat_wrapper,"id_mat","(int->mat)");
  install_function(diagonal_wrapper,"diagonal","(vec->mat)");
  install_function(stack_rows_wrapper,"stack_rows","([vec]->mat)");
  install_function(combine_columns_wrapper,"#","(int,[vec]->mat)");
  install_function(combine_rows_wrapper,"^","(int,[vec]->mat)");
  install_function(swiss_matrix_knife_wrapper,"swiss_matrix_knife"
      ,"(int,mat,int,int,int,int->mat)");
  install_function(swiss_matrix_knife_wrapper,"matrix slicer"
      ,"(int,mat,int,int,int,int->mat)"); // space make an untouchable copy
  
  install_function(gcd_wrapper,"gcd","(vec->int)");
  install_function(Bezout_wrapper,"Bezout","(vec->int,mat)");
  install_function(echelon_wrapper,"echelon","(mat->mat,mat,[int],int)");
  install_function(linear_solve_wrapper,"linear_solve","(mat,vec->|vec,int,mat)");
  install_function(diagonalize_wrapper,"diagonalize","(mat->vec,mat,mat)");
  install_function(adapted_basis_wrapper,"adapted_basis","(mat->mat,vec)");
  install_function(kernel_wrapper,"kernel","(mat->mat)");
  install_function(eigen_lattice_wrapper,"eigen_lattice","(mat,int->mat)");
  install_function(row_saturate_wrapper,"row_saturate","(mat->mat)");
  install_function(Smith_wrapper,"Smith","(mat->mat,vec)");
  install_function(invert_wrapper,"invert","(mat->mat,int)");
  install_function(section_wrapper,"mod2_section","(mat->mat)");
  install_function(subspace_normal_wrapper,
     "subspace_normal","(mat->mat,mat,mat,[int])");
 }

type_expr analyse_types(const expr& e,expression_ptr& p)
{ try
  { type_expr type; // this starts out as an |undetermined_type|
    p = convert_expr(e,type);
    return type;
  }
  catch (const type_error& err)
  { std::cerr << "Error during analysis of expression " << e.loc << std::endl;
    std::cerr << err.what() << ":\n  Subexpression " << err.offender
            << ' ' << err.offender.loc
            << "\n  has wrong type: found " << err.actual
            << " while " << err.required << " was needed.\n";

  }
  catch (const balance_error& err)
  { std::cerr << "Error in expression "
              << err.offender << ' ' << err.offender.loc << "\n  " 
              << err.what() ;
    for (auto it=err.variants.wcbegin(); not err.variants.at_end((it)); ++it)
      std::cerr << ( it==err.variants.wcbegin() ? ": { " : ", " ) << *it;
    std::cerr<< " }" << std::endl;
  }
  catch (const expr_error& err)
  { std::cerr << "Error in expression "
              << err.offender << ' ' << err.offender.loc << "\n  " 
              << err.what() << std::endl;
  }
  catch (const program_error& err)
  { std::cerr << "Error during analysis of expression " << e.loc << "\n  " 
              << err.what() << std::endl;
  }
  throw program_error("Expression analysis failed");

}

void global_set_identifier(const raw_id_pat &raw_pat, expr_p raw, int overload,
                          const source_location& loc)
{ do_global_set(id_pat(raw_pat),*expr_ptr(raw),overload,loc); }
  // ensure clean-up

void global_set_identifiers(raw_let_list d,const source_location& loc)
{ std::pair<id_pat,expr> pat_expr = zip_decls(d);
  do_global_set(std::move(pat_expr.first),pat_expr.second,1,loc);
}

definition_group::definition_group(unsigned int n_ids)
: bindings(), constness(n_ids)
{ bindings.reserve(n_ids); }

void definition_group::thread_bindings(const id_pat& pat,const type_expr& type)
{ if ((pat.kind & 0x1)!=0)
    add(pat.name,type.copy(),(pat.kind & 0x4)!=0);
  if ((pat.kind & 0x2)!=0)
    // recursively traverse sub-list for a tuple of identifiers
  { assert(type.kind()==tuple_type);
    wtl_const_iterator t_it(type.tuple());
    for (auto p_it=pat.sublist.begin(); not pat.sublist.at_end(p_it);
         ++p_it,++t_it)
      thread_bindings(*p_it,*t_it);
  }
}

void definition_group::add(id_type id,type_expr&& t, bool is_const=false)
{ for (auto it=bindings.cbegin(); it!=bindings.cend(); ++it)
  // check repeated identifiers
    if (it->first==id)
      throw program_error() << "Multiple occurrences of '"
        << main_hash_table->name_of(id) 
        << "' cannot be defined in same definition";

  if (t.kind()==function_type)
  { const auto& var=global_overload_table->variants(id);
    locate_overload(id,var,t.func()->arg_type);
    // may |throw|; otherwise ignore result
  }

  constness.set_to(bindings.size(),is_const);
  bindings.emplace_back(id,std::move(t));
}

void global_declare_identifier(id_type id, type_p t)
{ type_ptr saf(t); // ensure clean-up
  type_expr& type=*t;
  
  { unsigned int input_level = main_input_buffer->include_depth();
    *output_stream << std::setw(2*input_level) << "";
  }
  *output_stream << "Declaring identifier '" << main_hash_table->name_of(id) 
            << "': " << type << std::endl;
  static const shared_value undefined_value; // holds a null pointer
  global_id_table->add(id,undefined_value,std::move(type),false);
}

void clean_out_type_identifier(id_type id)
{
  const auto* defined_type = global_id_table->type_of(id);
  if (defined_type->raw_kind()!=tabled)
    return; // we cannot clear out the pro/in/jector functions, not recorded
  auto type_number = defined_type->type_nr();
  const auto& fields = type_expr::fields(type_number);
  if (not fields.empty())
  { if (defined_type->kind()==tuple_type)
      
      { for (unsigned i=0; i<fields.size(); ++i)
          if (fields[i]!=type_binding::no_id)
          { const auto* entry =
              global_overload_table->entry(fields[i],*defined_type);
            if (entry!=nullptr)
            { auto p = dynamic_cast<const projector_value*>(entry->value().get());
              if (p!=nullptr and p->id==fields[i] and p->position==i)
                global_overload_table->remove(fields[i],*defined_type);
            }
          }
      }
    else if(defined_type->kind()==union_type)
      
      { wtl_const_iterator comp_it (defined_type->tuple());
        for (unsigned i=0; i<fields.size(); ++i,++comp_it)
          if (fields[i]!=type_binding::no_id)
          { const auto* entry =
              global_overload_table->entry(fields[i],*comp_it);
            if (entry!=nullptr)
            { auto p = dynamic_cast<const injector_value*>(entry->value().get());
              if (p!=nullptr and p->id==fields[i] and p->position==i)
                global_overload_table->remove(fields[i],*comp_it);
            }
          }
      }
  }
}

void global_forget_identifier(id_type id)
{ if (global_id_table->is_defined_type(id))
    clean_out_type_identifier(id);
  bool was_known = global_id_table->remove(id);
  *output_stream << "Identifier '" << main_hash_table->name_of(id)
          << (was_known ? "' forgotten" : "' not known")
            << std::endl;
}

void global_forget_overload(id_type id, type_p t)
{ type_ptr saf(t); // ensure clean-up
  const type_expr& type=*t;
  const bool removed = global_overload_table->remove(id,type);
  *output_stream << "Definition of '" << main_hash_table->name_of(id)
            << '@' << type 
            << ( removed ? "' forgotten" : "' not known") << std::endl;
}

void type_define_identifier
  (id_type id, type_p t, raw_id_pat ip, const source_location& loc)
{ type_ptr saf(t); id_pat field_pat(ip); // ensure clean-up
  type_expr& type=*t;
  const auto& fields = field_pat.sublist;
  const auto n=length(fields);
  definition_group group(n);
  std::vector<shared_function> jectors; jectors.reserve(n);
  std::vector<id_type> names(n,id_type(type_binding::no_id));
  try
  {
    if (not fields.empty()) // do this before we move from |type|
      
      { assert(type.kind()==tuple_type or type.kind()==union_type);
        auto tp_it =wtl_const_iterator(type.tuple());
        auto id_it=fields.wcbegin();
        if (type.kind()==tuple_type)
        { for (unsigned i=0; i<n; ++i,id_it++,tp_it++)
            if (id_it->kind==0x1) // field selector present
            { names[i]=id_it->name;
              jectors.push_back
                (std::make_shared<projector_value>(type,i,names[i],loc));
              group.add(names[i],type_expr(type.copy(),tp_it->copy()));
                // projector type
            }
        }
        else
        { for (unsigned i=0; i<n; ++i,id_it++,tp_it++)
            if (id_it->kind==0x1) // field selector present
            { names[i]=id_it->name;
              jectors.push_back
                (std::make_shared<injector_value>(type,i,names[i],loc));
              group.add(names[i],type_expr(tp_it->copy(),type.copy()));
                // injector type
            }
        }
      }

    
    { for (auto it=group.begin(); it!=group.end(); ++it)
        if (it->first==id)
          throw program_error()
            << "Type definition of '" << main_hash_table->name_of(id) 
                    << "' cannot contain a field of the same name";
    
      bool redefine = global_id_table->is_defined_type(id);
      if (not redefine)
        
        { const bool p = global_id_table->present(id) and
                         not global_id_table->is_defined_type(id);
          if (p or not global_overload_table->variants(id).empty())
            throw program_error()
               << "Cannot define '" << main_hash_table->name_of(id) 
               << "' as a type; it is in use as " 
               << (p? "global variable" : "function");
        }
      else // those tests were already done when |id| became a type identifier
        clean_out_type_identifier(id);
      
      { unsigned int input_level = main_input_buffer->include_depth();
        *output_stream << std::setw(2*input_level) << "";
      }
      *output_stream << "Type name '" << main_hash_table->name_of(id) 
                << (redefine ? "' redefined as " : "' defined as ") << type
                << std::endl;
    }
    if (group.begin()!=group.end())
      // only need the following when there are field names
      
      { 
        { unsigned int input_level = main_input_buffer->include_depth();
          *output_stream << std::setw(2*input_level) << "";
        }
      *output_stream << "  with " << (type.kind()==tuple_type ? "pro" : "in");
        for (auto it=group.begin(); it!=group.end(); ++it)
        { global_overload_table->add
            (it->first,std::move(jectors[it-group.begin()]),std::move(it->second));
          *output_stream << (it==group.begin() ? "jectors: " : ", ")
                       << main_hash_table->name_of(it->first);
        }
        *output_stream << '.' << std::endl;
      }

    global_id_table->add_type_def(id,std::move(type));

  }
  catch (program_error& err)
  { std::string mes; mes.swap(err.message);
    throw err << "Error in type definition " << loc << ":\n"  << mes
              << "\n  Type definition aborted";
  }
}

void process_type_definitions (raw_typedef_list l, const source_location& loc)
{ typedef_list defs(l);
  defs.reverse(); // since the parser collects by prepending new nodes
  const auto old_size = type_expr::table_size(); // for roll back
  try
  {
    static const type_nr_type absent = -1;
    std::vector<type_nr_type> translate (main_hash_table->nr_entries(),absent);
  
  {
    type_nr_type count = type_expr::table_size();
    for (auto it=defs.begin(); not defs.at_end(it); ++it,++count)
      if(it->id!=type_binding::no_id)
      { id_type id = it->id;
        
        { const bool p = global_id_table->present(id) and
                         not global_id_table->is_defined_type(id);
          if (p or not global_overload_table->variants(id).empty())
            throw program_error()
               << "Cannot define '" << main_hash_table->name_of(id) 
               << "' as a type; it is in use as " 
               << (p? "global variable" : "function");
        }
        if (translate[id]==absent)
          translate[id] = count;
        else
          throw program_error()
            << "Repeated definition of '"  << main_hash_table->name_of(id);
      }
    for (auto it=defs.begin(); not defs.at_end(it); ++it)
      for (auto jt=it->fields.wcbegin(); not jt.at_end(); ++jt)
        if ((jt->kind&0x1)!=0 and translate[jt->name]!=absent)
        throw program_error()
          << "Used '" << main_hash_table->name_of(jt->name) 
          << "' as defined type AND as field name";
  }
  
  { containers::stack<type_p> work;
    for (auto it=defs.begin(); not defs.at_end(it); ++it)
      work.push(it->type);
    while (not work.empty())
    { auto& t = *work.top();
      work.pop(); // hold pointer as non-owned reference; drop the pointer
      switch(t.raw_kind())
      { default: break;
      case function_type:
      { auto f=t.func();
          work.push(&f->result_type);
          work.push(&f->arg_type);
        }
      break;
      case row_type: work.push(t.component_type());
      break;
      case tuple_type: case union_type:
        for (wtl_iterator it(t.tuple()); not it.at_end(); ++it)
          work.push(&*it);
      break;
      case tabled: // this case is what it is all about
        { id_type id = t.type_nr();
   // narrows the integer, but it was really an identifier code
          if (translate[id]!=absent)
            t = type_expr(translate[id]); // replace by future tabled reference
          else if (global_id_table->is_defined_type(id))
            t = global_id_table->type_of(id)->copy();
          else
            throw program_error()
              << "Type identifier '" << main_hash_table->name_of(id) 
              << "' does not refer to any type";
        }
      break;
      }
    }
  }

    std::vector<std::pair<id_type,const_type_p> > b; b.reserve(length(defs));
    for (auto it=defs.begin(); it!=end(defs); ++it)
      b.emplace_back(it->id,it->type);
    auto type_nrs = type_expr::add_typedefs(b);

    
    { unsigned int i = 0; // position within |defs|
      containers::sl_list<definition_group> store;
      for (auto it=defs.begin(); not defs.at_end(it); ++it,++i)
        if (not it->fields.empty())
        { const auto& fields = it->fields;
          const auto& type = type_expr(type_nrs[i]);
          
          { assert(type.kind()==tuple_type or type.kind()==union_type);
            auto record = store.end(); // iterator that will point to next pushed item
            store.emplace_back(definition_group(length(fields)));
          
            auto tp_it =wtl_const_iterator(type.tuple());
            if (type.kind()==tuple_type)
            {
              for (auto id_it=fields.wcbegin(); not fields.at_end(id_it);
                   ++id_it,++tp_it)
                if (id_it->kind==0x1) // field selector present
                  record->add(id_it->name,type_expr(type.copy(),tp_it->copy()));
                    // projector type
            }
            else
            {
              for (auto id_it=fields.wcbegin(); not fields.at_end(id_it);
                   ++id_it,++tp_it)
                if (id_it->kind==0x1) // field selector present
                  record->add(id_it->name,type_expr(tp_it->copy(),type.copy()));
                    // injector type
            }
          }
        }
    
      auto store_it = store.wbegin(); // rewind the list of field lists
      for (auto it=(i=0,defs.wcbegin()); not defs.at_end(it); ++it,++i)
      {
        const auto& fields = it->fields;
        const auto type_nr = type_nrs[i];
        const auto& type = type_expr(type_nr);
        if (it->id!=type_binding::no_id)
        {
          if (global_id_table->is_defined_type(it->id))
            clean_out_type_identifier(it->id);
          global_id_table->add_type_def(it->id,type_expr(type_nr));
        }
        
        { unsigned int input_level = main_input_buffer->include_depth();
          *output_stream << std::setw(2*input_level) << "";
        }
        if (it->id==type_binding::no_id)
          *output_stream << "Anonymous type "
                         << type << std::endl;
        else
          *output_stream << "Type name '" << main_hash_table->name_of(it->id) 
            << "' defined as " << type.untabled() << std::endl;
        if (not fields.empty())
        { auto& group = *store_it;
          
          {
            bool tup = type.kind()==tuple_type;
            
            { unsigned int input_level = main_input_buffer->include_depth();
              *output_stream << std::setw(2*input_level) << "";
            }
          *output_stream << "  with " 
             << (tup ? "pro" : "in");
            auto group_it = group.begin(); unsigned int j=0;
            std::vector<id_type> names(length(it->fields),id_type(type_binding::no_id));
            for (auto id_it=fields.wcbegin(); not fields.at_end(id_it); ++id_it,++j)
            {
              *output_stream << (j==0 ? "jectors: " : ", ");
              if (id_it->kind==0x1) // field selector present
              { auto name = names[j] = id_it->name; assert(name==group_it->first);
                shared_function tor; // function object to store
                if (tup)
                  tor = std::make_shared<projector_value>(type,j,name,loc);
                else
                  tor = std::make_shared<injector_value>(type,j,name,loc);
                global_overload_table->add 
                  (name,std::move(tor),std::move(group_it->second));
                *output_stream << main_hash_table->name_of(name);
                ++group_it; // advance only here
              }
            }
            *output_stream << '.' << std::endl;
          
            type_expr::set_fields(type_nrs[i],std::move(names));
          
          }
         ++store_it; // |store_it| only advances when fields were present
        }
      }
    }
  }
  catch (program_error& err)
  { type_expr::reset_table_size(old_size); // roll back any extension made
    std::string mes; mes.swap(err.message);
    throw err << "Error in 'set_type' command " << loc << ":\n"  << mes
              << "\n  Type definition aborted";
  }
}

void type_of_expr(expr_p raw)
{ expr_ptr saf(raw); const expr& e=*raw;
  try
  { expression_ptr p;
    *output_stream << "Type: " << analyse_types(e,p) << std::endl;
  }
  catch (std::exception& err) { std::cerr<<err.what()<<std::endl; }
}

void type_of_type_name(id_type id)
{ const auto* type = global_id_table->type_of(id);
  const std::vector<id_type>* fields = nullptr;
  if (type->raw_kind()==tabled)
    fields = &type_expr::fields(type->type_nr());
  *output_stream << "Defined type: ";
  if (fields==nullptr or fields->empty())
    *output_stream << type->untabled();
  else
  { char sep = type->kind()==tuple_type ? ',' : '|';
    auto f_it = fields->begin();
    for (wtl_const_iterator it(type->tuple()); not it.at_end(); ++it,++f_it)
      *output_stream << (f_it==fields->begin() ? '(' : sep)
       << ' ' << *it << ' ' 
       << (*f_it == type_binding::no_id ? "." : main_hash_table->name_of(*f_it))
       << ' ';
    *output_stream << ')';
  }
  *output_stream << std::endl;
}

void show_overloads(id_type id,std::ostream& out)
{ const overload_table::variant_list& variants =
   global_overload_table->variants(id);
   out
   << (variants.empty() ? "No overloads for '" : "Overloaded instances of '")
 << main_hash_table->name_of(id) << '\'' << std::endl;
 for (std::size_t i=0; i<variants.size(); ++i)
   out << "  "
    << variants[i].type().arg_type << "->" << variants[i].type().result_type 
    << std::endl;
}

void show_ids(std::ostream& out)
{ out << "Overloaded operators and functions:\n"
      << *global_overload_table 
      << "Global variables:\n" << *global_id_table;
}

void vector_value::print(std::ostream& out) const
{ auto l=val.size(); std::size_t w=0; std::vector<std::string> tmp(l);
  for (std::size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (std::size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
}

void rational_vector_value::print(std::ostream& out) const
{ auto l=val.size(); std::size_t w=0; std::vector<std::string> tmp(l);
  for (std::size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val.numerator()[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (std::size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
  out << '/' << val.denominator();
}

void matrix_value::print(std::ostream& out) const
{ auto k=val.numRows(),l=val.numColumns();
  if (k==0 or l==0)
  {  out << "The " << k << 'x' << l << " matrix"; return; }
  std::vector<std::size_t> w(l,0);
  for (std::size_t i=0; i<k; ++i)
    for (std::size_t j=0; j<l; ++j)
    { std::ostringstream s; s<<val(i,j); auto len=s.str().length();
      if (len>w[j]) w[j]=len;
    }
  out << std::endl << std::right;
  for (std::size_t i=0; i<k; ++i)
  { out << '|';
    for (std::size_t j=0; j<l; ++j)
      out << std::setw(w[j]+1) << val(i,j) << (j<l-1 ? ',' : ' ');
    out << '|' << std::endl;
  }
}

int_Vector row_to_vector(const row_value& r)
{ int_Vector result(r.val.size());
  for(std::size_t i=0; i<r.val.size(); ++i)
    result[i]=force<int_value>(r.val[i].get())->int_val();
  return result;
}

own_row vector_to_row(const int_Vector& v)
{ own_row result = std::make_shared<row_value>(v.size());
  for(std::size_t i=0; i<v.size(); ++i)
    result->val[i]=std::make_shared<int_value>(v[i]);
  return result;
}

void install_function
 (wrapper_function f,const char*name, const char* type_string)
{ type_ptr type = mk_type(type_string);
  std::ostringstream print_name; print_name<<name;
  if (type->kind()!=function_type)
    throw logic_error
     ("Built-in with non-function type: "+print_name.str());
  print_name << '@' << type->func()->arg_type;
  auto val = std::make_shared<builtin_value<false> >(f,print_name.str());
  global_overload_table->add
    (main_hash_table->match_literal(name),std::move(val),std::move(*type));
}

void sizeof_string_wrapper(expression_base::level l)
{ auto s=get<string_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void sizeof_vector_wrapper(expression_base::level l)
{ auto s=get<vector_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void sizeof_ratvec_wrapper(expression_base::level l)
{ auto s=get<rational_vector_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void matrix_ncols_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<int_value>(m->val.numColumns()));
}

void mm_prod_wrapper(expression_base::level l)
{ shared_matrix rf=get<matrix_value>(); // right factor
  shared_matrix lf=get<matrix_value>(); // left factor
  check_size(lf->val.numColumns(),rf->val.numRows());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lf->val*rf->val));
}

 void transpose_mat_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(m->val.transposed()));
}

void id_mat_wrapper(expression_base::level l)
{ int i=get<int_value>()->int_val();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (int_Matrix(std::max(i,0)))); // identity
}

} }

